#include <stdlib.h>
#include <tos.h>
#include <errno.h>
#include "hw-types.h"
#include <stdio.h>
#include <string.h>
#include <time.h>

/*#define ERR2FILE */

#define SOCK_STREAM		1L
#define SOCK_DGRAM		2L
#define SOCK_RAW			3L
#define SOCK_RDM			4L
#define SOCK_SEQPACKET	5L	
  	
#define 	AF_INET			2	/* siehe socket.h bei in2cat */ 
 
#define INET_VERSION	0x20071111l	/* Original aus inet.h bernommen! */
#define OVL_REVISION 0x20030206ul

/* unique header identification */
/* Use which ever version you feel more comfortable with,
   or more appropriately which ever version your compiler is 
   more comfortable with
*/
#define OVL_MAGIC	0x57695265L		/* "WiRe" 	 */

#define OF_SIMPLE  0x0000uL /* nothing special to do */
#define OF_STARTUP 0x0001uL /* needs to get main() started, Pexec(4) */
#define OF_CHLDPRC 0x0002uL /* needs to be started as process, Pexec(104) */
#define OF_THREAD  0x0003uL

#define FTAB_SIMPLE  0x0000uL /* standard FTAB interface   */
#define FTAB_NETWORK 0x0001uL /* networking FTAB interface */	

typedef struct in_addr 
{
	union 
	{
		struct 
		{
			unsigned char s_b1,s_b2,s_b3,s_b4;
		} S_un_b;
		struct 
		{
			unsigned int s_w1,s_w2;
		} S_un_w;
		unsigned long S_addr;
	} S_un;
} in_addr;
						
typedef struct 
{
	int	sin_family;
	unsigned int sin_port;
	struct in_addr sin_addr;
	char sin_zero[8];
} SOCKADDR_IN;

typedef struct fd_set
{
	unsigned long fd_count;       /* how many are SET? */
   unsigned long fd_array[64];   /* an array of SOCKETs */
} fd_set;

typedef struct
{
	long    tv_sec;         /* seconds */
   long    tv_usec;        /* and microseconds */
}timeval;
	
						
typedef struct MERKE 	
{
	void 	*sptr;
	unsigned long	h, wh;
	struct MERKE *next;
	struct MERKE *prev;
	void *s_in;
} MERKE;

/*--- OVL information ---*/
struct ovl_info_t
{
	char *name;
	char *version;
	char *date;
	char *author;
};

typedef struct {
	int cdecl (*host_addr) (const char * host, long * addr);
	long  cdecl (*connect) (long addr, long port, long tout_sec);
	long  cdecl (*send)    (long fh, const char * buf, size_t len);
	long  cdecl (*recv)    (long fh, char       * buf, size_t len);
	void  cdecl (*close)   (long fh);
	long  cdecl (*instat)  (long fh);
	long  cdecl (*select)  (long timeout, long * rfds, long * wfds);	
	const char * cdecl (*info) (void);
} INET_FTAB;


/*--- Module functions used by the client ---*/
typedef struct ovl_methods_t
{
	long magic;    /* must be "WiRe" */
	long revision; /* interface revision date in ISO yyyymmdd  BCD */
	long flags;    /* startup basic information: */
	long ftabtype; /* Function Table mapping reference */
	long                cdecl (*ovl_init)   (void);
	struct ovl_info_t * cdecl (*ovl_version)(void);
	void              * cdecl (*ovl_getftab)(void);
	long                cdecl (*ovl_free)   (void);
	long * magic_l;
} OVL_METH;

typedef	void (*void_fun_pointer) (void);

/* Funktionen aus der in2cat.dll */
extern long win_sock_init(void);
extern void *win_socket(void);
extern long win_create(void* sock,long type);
extern unsigned long win_lstErr(void* sock);
extern unsigned long win_gethostbyname(char *name);
extern long win_connect(void* sock, const void *addr, long addrlen);
extern long win_connectTHS(void* DiaSt, const void *addr, long addrlen);
extern long win_connectC(void* sock, char *url, unsigned long port);
extern long win_connectCTHS(void* DiaSt, char *url, unsigned long port);
extern long win_send(void* s, const void *msg, long len, long flags);
extern long win_sendTHS(void* DiaSt, const void *msg, long len, long flags);
extern long win_recvfrom(void* sock, void *buf, long len, long flags, void *from, long *fromlen, long timeOut);
extern long win_recvfromTHS(void *DiaSt, void *buf, long len, long flags, void *from, long *fromlen, long timeOut);
extern long win_Receive(void* sock, void *lpBuf, long nBuflen, long nFlags);
extern long win_ReceiveTHS(void* DiaSt, void *lpBuf, long nBuflen, long nFlags);
extern long win_sclose(void* sock);
extern unsigned long win_getHandle(void *sock);
extern long win_select(void* sock, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, timeval *timeout);
extern long win_done(void *DiaSt, long *ret);
extern long dll_version(void);
extern long dll_versionAbwC(void);

/*
 set of functions which are always defined in an OVL 
long cdecl myovl_init(void);
long cdecl myovl_free(void);
struct ovl_info_t *cdecl myovl_version(void);
void * cdecl myovl_getftab(void);

 "Netz"-Funktionen des Overlays 
int cdecl ovl_host_addr(const char * host, long * addr);
long cdecl ovl_connect(long addr, long port, long tout_sec);
long cdecl ovl_send(long fh, const char * buf, size_t len);
long cdecl ovl_recv(long fh, char * buf, size_t len);
long cdecl ovl_instat(long fh);
long cdecl ovl_select(long timeout, long * rfds, long * wfds);
void cdecl ovl_close(long fh);
const char * cdecl ovl_info (void);									
*/

/* Globale Variablen */
MERKE 	*firstmerk;
FILE 		*fdebug;
int		first_error;
int		has_threads = 0;
void		*old_term;

/*============================================================================*/

int Bconws(int device, char *str)
{
	int	cnt = 0;
	int	odev	=	device;
	
	if (odev == 3)
		odev = 4;
	else
		if (odev == 4)
			odev = 3;
	
	while (*str)
	{
		cnt++;
		
		while (!Bcostat(odev));
		
		if (*str == '\n')
		{
			Bconout(device, '\r');
			cnt++;
		}
		
		Bconout(device, *str++);
	}		
	
	return cnt;
}

void new_term(void)
{
	/* char	s[256]; */
	MERKE	*freeptr, *currptr;
	void (*terminate) (void) = old_term;
	
	/* Bconout(2, 65); */	
	
	Setexc(0x102, (void_fun_pointer) old_term);

	currptr = firstmerk;
	firstmerk = NULL;
	
	/* if (currptr == NULL)
		Bconws(2, "firstmerk = NULL\n"); */
	
	while (currptr)
	{
		/*Bconws(2, "Freizugebender currptr->sptr = ");
		Bconws(2, ltoa((long) currptr->sptr, s, 10));
		Bconws(2, "\n");*/	
	
		freeptr = currptr;
		currptr = currptr->next;
		
		win_sclose(freeptr->sptr);
	}
	
	/* Bconout(2, 90); */
	
	terminate();
}

/*============================================================================*/

struct ovl_info_t ovl_data = {
	"WinOVL",
	"0.2", __DATE__ ", " __TIME__, "Uwe & Jrgen Holtkamp"
};	

/*============================================================================*/
struct ovl_info_t *cdecl myovl_version(void)
{
	return((struct ovl_info_t *)&ovl_data);
}

/*============================================================================*/

const char * cdecl ovl_info (void)
{
	return "Windows Socket";
}

/*============================================================================*/

#ifdef ERR2FILE
void add_statemsg(char *s)
{
	if (first_error)
	{
		fdebug = fopen("C:\\OVLERROR.TXT", "w");
		first_error = 0;
	}
	else
		fdebug = fopen("C:\\OVLERROR.TXT", "a");
		
	if (fdebug)
	{
		fprintf(fdebug, s);
		fclose(fdebug);
	}
}
#else
void add_statemsg(char *s)
{
	printf(s);
}
#endif	

/*============================================================================*/

MERKE *get_entry(unsigned long fh, int atari)
{
	MERKE		*currmerk;
	char		found;
	
	currmerk = firstmerk;
	found = 0;
	
	while ((currmerk != NULL) && (!found)) 
	{
		/*found = currmerk->h == fh;*/
		found = ((currmerk->h == fh) && (atari)) || ((currmerk->wh == fh) && (!atari));
		if (!found) 
			currmerk = currmerk->next;
	}
	if (found) 
		return currmerk;
	else
	{
		return NULL;
	}
}	

/*============================================================================*/

void free_merklist(void)
{
	MERKE	*freeptr, *currptr;

	currptr = firstmerk;
	firstmerk = NULL;
	
	while (currptr)
	{
		freeptr = currptr;
		currptr = currptr->next;
		
		win_sclose(freeptr->sptr);
		
		Mfree(freeptr->s_in);
		Mfree(freeptr);
	}
}

/*============================================================================*/

int close_and_cut(MERKE *currmerk)
{
	if (currmerk == NULL)
		return 0;
	else
	{
		win_sclose(currmerk->sptr);
		
		if (currmerk->prev)
			currmerk->prev->next = currmerk->next;
			
		if (currmerk->next)
			currmerk->next->prev = currmerk->prev;
			
/*		if (currmerk == firstmerk)	
			firstmerk = NULL; */
		
		if (currmerk == firstmerk)
			firstmerk = currmerk->next;
		
		Mfree(currmerk->s_in);
		Mfree(currmerk);
		return 1;
	}
}

/*============================================================================*/

unsigned int htons(unsigned int w)
{
	union 
	{
		struct
		{
			unsigned char b1, b2;
		} b;
		unsigned int w;
	} ein, aus;
	
	
	ein.w = w;
	aus.b.b1 = ein.b.b2;
	aus.b.b2 = ein.b.b1;

	return aus.w;
}


/*============================================================================*/
	
long cdecl myovl_init(void)
{
	/* ovl_init() called by client to initialize ovl */
	first_error = 1;
	firstmerk = NULL;

	has_threads = (dll_version() >= 211L) && (dll_versionAbwC() <= 211L);

	(long) old_term = 0L;
	
	if (win_sock_init() == 0L)
	{
		add_statemsg("Fehler bei Initialisierung der Windows-Socket-Library! \n");
		return -1L;
	}	
	else
	{
		if (has_threads)
			old_term = Setexc(0x102, new_term);
		
		return INET_VERSION;	/* Rckgabewert = Versionsnummer = Datum der Version */
	}
}

/*============================================================================*/

/* host reicht doch als Pointer, oder? */
int cdecl ovl_host_addr(const char *host, long *addr)  
{
	char	s[256];

	strcpy(s, host); 

	(unsigned long) *addr = win_gethostbyname(s);
	
	if (*addr)
		return 0;
	else
		return -1;
}	

/*============================================================================*/

unsigned int loword(long l)
{
	return (unsigned int) (l & 0xffff);
}

/*============================================================================*/

void wait_for_a_while(long ms)
{
	clock_t	timeout;

	timeout = clock() + ms / 5L;
		
	
	do
		Syield();
	while (clock() < timeout);
}

/*============================================================================*/

static BOOL sig_tout = FALSE;

static void sig_alrm (long sig)
{
	(void)sig;
	sig_tout = TRUE;
}

/*============================================================================*/

long thread_connect(void *DiaSt, const void *addr, long addrlen)
{
	long retL;
	
	if (has_threads)
	{
		if (win_connectTHS(DiaSt, addr, addrlen))
		{
			add_statemsg("Fehler beim Erzeugen des Threads bei win_connectTHS\n");
			return -1L;
		}
		else
		{
			while((win_done(DiaSt, &retL) == 0) && (!sig_tout))
				wait_for_a_while(10L);
				
			retL = sig_tout ? -ETIMEDOUT : retL;
			
			return retL;
		}
	}
	else
		return win_connect(DiaSt, addr, addrlen);
}

long cdecl ovl_connect(long addr, long port, long tout_sec)
{

	void				*sockptr;
	unsigned long	i;
	long				result;
	MERKE				*currmerk, *newmerk;
	SOCKADDR_IN		*s_in;
	void				*alrm;
	long				connect_ret;

	result = -1L;
	
	sockptr = win_socket();
	if (sockptr == NULL) 
		add_statemsg("Fehler bei der Speicherallozierung der Socket!\n");
	else
	{
		if (win_create(sockptr, SOCK_STREAM) == 0L)
			add_statemsg("Fehler bei der Create-Funktion des Socket-Objekts!\n");
		else
		{
			/* Verbindung zum Server herstellen */
			s_in = (SOCKADDR_IN *) Malloc(sizeof(SOCKADDR_IN));
			if (s_in != NULL)
			{
				s_in->sin_family = AF_INET;  
				s_in->sin_port = htons(loword(port));		
				s_in->sin_addr = *(struct in_addr *)&addr;
				
				/* Vor dem Aufruf des eigentlichen Connects Timeout-Signalhandler setzen */
				alrm = Psignal(14, sig_alrm);	/* bei Signal "SIGALRM" eigenen Handler aufrufen */
				if ((long) alrm >= 0)	/* Psignal war erfolgreich */
				{
					sig_tout = FALSE;	/* Initialisierung: noch kein Timeout eingetroffen */
					Talarm(tout_sec); /* nach tout_sec Sekunden gibt es das SIGALRM-Signal */
				}
			
				connect_ret = thread_connect(sockptr, s_in, sizeof(SOCKADDR_IN));
				
				/* Timeout-Signalhandler bzw. SIGALRM wieder zurcksetzen */
				if ((long) alrm >= 0)
				{
					Talarm(0L);
					Psignal(14, alrm);
				}
				
				if (connect_ret)
				{
					add_statemsg("Fehler beim Connect!\n");
					win_sclose(sockptr);
					
					Mfree(s_in);
					
					if (connect_ret == -ETIMEDOUT)
						result = -ETIMEDOUT;
				}
				else
				{
					/* Handle und dazugehrende Socket dynamisch merken */
					newmerk = Malloc(sizeof(MERKE));
					
					if (newmerk)
					{
						newmerk->sptr = sockptr;
						newmerk->wh = win_getHandle(sockptr);
						newmerk->s_in = s_in;
						
						i = 0L;
						
						if (firstmerk == NULL)
						{
							i++;
							firstmerk = newmerk;
							newmerk->prev = NULL;
							newmerk->next = NULL;
						}
						else
						{
							i++;
							currmerk = firstmerk;
							
							/* Finde ersten Eintrag, der nicht der Sortierreihenfolge der handles (1, 2, 3...)
								entspricht. Falls gefunden: fge neuen Eintrag davor ein; falls nicht gefunden:
								fge neuen Eintrag am Ende ein */
								
							while ((currmerk->next != NULL) && (currmerk->h == i))
							{
								currmerk = currmerk->next;
								i++;
							}
							
							/* Entweder ist hier currmerk->next == NULL oder currmerk->h != i */
							if (currmerk->h != i)
							{
								/* Kette muss aufgebrochen werden! */
								if (currmerk->prev)
									currmerk->prev->next = newmerk;	
								else
									firstmerk = newmerk;
								newmerk->prev = currmerk->prev;
								newmerk->next = currmerk;
								currmerk->prev = newmerk;
							}
							else
							{
								/* Es muss am Ende der Kette alloziert werden! */
								i++;
								newmerk->prev = currmerk;
								currmerk->next = newmerk;
								newmerk->next = NULL;
							}						
						}
						
						newmerk->h = i;
						
						/* Alles hat geklappt: handle zurckgeben */
						result = (long) newmerk->h;
					}
					else
					{
						win_sclose(sockptr);

			 			add_statemsg("Fehler beim Connect: Speichermangel!\n");
						
						Mfree(s_in);
					};				
				}
			}
			else
				add_statemsg("Fehler beim Allozieren der SOCKADDR_IN-Struktur! Kein Connect mglich!\n");
		}
	}
	
	return result;
}

/*============================================================================*/

long thread_send(void* DiaSt, const void *msg, long len, long flags)
{
	long 	retL;
		
	if (has_threads)
	{
		if (win_sendTHS(DiaSt, msg, len, flags))
		{
			add_statemsg("Fehler beim Erzeugen des Threads bei win_sendTHS\n");
			return -1L;
		}
		else
		{
			while(win_done(DiaSt, &retL) == 0)
				wait_for_a_while(10L);
				
			return retL;
		}	
	}
	else
		return win_send(DiaSt, msg, len, flags);
}

long cdecl ovl_send(long fh, const char *buf, size_t len)
{
	MERKE	*currmerk;
	long	result = -1L;
	/* int	h; */
	
	/* Finde den sockptr, der zum Handle fh gehrt */
	currmerk = get_entry((unsigned long) fh, 1);
	
	if (currmerk)
	{
		result = thread_send(currmerk->sptr, buf, len, 0L);
		if (result < 0)
			add_statemsg("Fehler beim Senden!\n");
	}
	else
		add_statemsg("Der fr das Senden ntige Socketpointer konnte nicht gefunden werden!\n");	
		
	return result;
}

/*============================================================================*/

long thread_recvfrom(void *DiaSt, void *buf, long len, long flags, void *from, long *fromlen, long timeOut)
{
	long retL;
		
	if (has_threads)
	{	
		if (win_recvfromTHS(DiaSt, buf, len, flags, from, fromlen, timeOut))
		{
			add_statemsg("Fehler beim Erzeugen des Threads bei win_recvfromTHS\n");
			return -1L;
		}
		else
		{
			while(win_done(DiaSt, &retL) == 0)
				wait_for_a_while(10L);
				
			return retL;
		}		
	}
	else
		return win_recvfrom(DiaSt, buf, len, flags, from, fromlen, timeOut);
}


long thread_Receive(void* DiaSt, void *lpBuf, long nBuflen, long nFlags)
{
	long retL;
	
		
	if (has_threads)
	{
		if (win_ReceiveTHS(DiaSt, lpBuf, nBuflen, nFlags))
		{
			add_statemsg("Fehler beim Erzeugen des Threads bei win_ReceiveTHS\n");
			return -1L;
		}
		else
		{
			while(win_done(DiaSt, &retL) == 0)
				wait_for_a_while(10L);

			return retL;
		}		
	}
	else
		return win_Receive(DiaSt, lpBuf, nBuflen, nFlags);
}

long cdecl ovl_recv(long fh, char *buf, size_t len) /* len eig. ulong! */
{
	long		ldummy;
	MERKE		*currmerk;
	long		result = 0L, n;
	
	/* Finde den sockptr, der zum Handle fh gehrt */
	currmerk = get_entry((unsigned long) fh, 1);
	
	if (currmerk) 
	{
		ldummy = 0L;	/* vmtl. unntig */
		
		do
		{
			/* Empfangsbefehl mit Timeout von 3 Sekunden (letzter Parameter) */
			/* Ohne Timeout hngt Highwire z. B. bei www.facebook.com. */
			n = thread_recvfrom(currmerk->sptr, buf, len, 0L, NULL, &ldummy, 3000L);
			/*n = thread_Receive(currmerk->sptr, buf, len, 0L);*/
			if (n > 0)
			{
				result = result + n;
				(long) buf = (long) buf + n;
				len = len - n;
			}
			else
			{
				if ((n <= 0L) && (result == 0L))
				{
					if (n < 0L)
					{
						result = n;
					}
					else
					{
						/* n = 0: Server liefert keine Daten mehr */
						result = -316L;	/* ECONNRESET = Connection reset by peer */
					}
				}
			}
		} while (!((len == 0L) || (n <= 0L)));	/* Fehler: n < 0; fertig oder Sendeabbruch: n = 0 */
	}
	else
		add_statemsg("Der fr den Empfang ntige Socketpointer konnte nicht gefunden werden!\n");
		
	return  result;
}

/*============================================================================*/

void cdecl ovl_close(long fh)
{
	if (fh > 0)
	{
		if (!close_and_cut(get_entry((unsigned long) fh, 1)))
			add_statemsg("Der zum Schlieen der Socket ntige Socketpointer konnte nicht gefunden werden!\n");
	}
}

/*============================================================================*/

long cdecl ovl_instat(long fh)
{
	long ret = -1L;
	char buf[1];
	MERKE *currmerk;
	
	/* Finde den sockptr, der zum Handle fh gehrt */
	currmerk = get_entry((unsigned long) fh, 1);
	
	if (currmerk) 
	{	
		ret = thread_Receive(currmerk->sptr, buf, 1, 2);
		if /*(ret == 0L) ||*/ (ret == 10054L)	/* 10054 = WSAECONNRESET */
		{
			ret = -316L;	/* ECONNRESET = Connection closed */
		}
	}
	
	return ret;
}

/*============================================================================*/

long cdecl ovl_select(long timeout, long * rfds, long * wfds)
{
	long ret = 0L;
	fd_set 	*readfds, *writefds;
	int		i, bit;
	MERKE		*currmerk;
	timeval	timerec;
	long		mask;
	clock_t	uhrzeit;
	
	/* Speicher fr die fd_set-Struktur anfordern */
	readfds = Malloc(sizeof(fd_set));
	writefds = Malloc(sizeof(fd_set));
	
	if ((readfds) && (writefds))
	{
		/* Beide Strukturen initialisieren */
		readfds->fd_count = 0L;
		writefds->fd_count = 0L;
		
		for (i = 0; i < 64; i++)
		{
			readfds->fd_array[i] = 0L;
			writefds->fd_array[i] = 0L;
		} /* memset(readfds->fd_array, 0, 64 * sizeof(long)) ?? */
	
		/* Bitfeld rfds nur auslesen, wenn es existiert und ungleich Null ist */
		if (rfds && *rfds)
		{
			for (bit = 0; bit < 32; bit++)
			{
				mask = 1 << bit;
				
				if (*rfds & mask)
				{
					currmerk = get_entry((unsigned long) bit, 1);
					
					if (currmerk)
					{
						/* Es gibt eine (geffnete) Socket mit dem Atarihandle "bit" */
						readfds->fd_array[readfds->fd_count] = currmerk->wh;
						readfds->fd_count++;
					}
					else
					{
						add_statemsg("Ein Bit im Lesefeld des Selectaufrufes reprsentiert eine nicht geffnete Socket!/n");
					}
				}
			}
		}
		
		/* Bitfeld wfds nur auslesen, wenn es existiert und ungleich Null ist */
		if (wfds && *wfds)
		{
			for (bit = 0; bit < 32; bit++)
			{
				long mask = 1 << bit;
				
				if (*wfds & mask)
				{
					currmerk = get_entry((unsigned long) bit, 1);
					
					if (currmerk)
					{
						/* Es gibt eine (geffnete) Socket mit dem Atarihandle "bit" */
						writefds->fd_array[writefds->fd_count] = currmerk->wh;
						writefds->fd_count++;
					}
					else
					{
						add_statemsg("Ein Bit im Schreibfeld des Selectaufrufes reprsentiert eine nicht geffnete Socket!/n");
					}
				}
			}
		}

		/* Lnger als 0,5 Sekunden sollte MagiC nicht "stehen"! */
		timerec.tv_sec  = (long)(min(timeout, 500) /1000L);    /* calc seconds from milliseconds */
		timerec.tv_usec = ((min(timeout, 500)%1000L)*1000L); /* calc remainder in microsecs */ 
		
		uhrzeit = clock() + (clock_t) (timeout * CLK_TCK / 1000L);
			
		while (clock() < uhrzeit)
		{
			ret = win_select(NULL, readfds, writefds, NULL, &timerec);
			
			if (ret)
				break;
			for (i = 0; i < 100; i++)
				Syield();
		}
/*		
		if (timeout > 500)	/* Lnger als 0,5 Sekunden sollte MagiC nicht "stehen"! */
		{
			timerec.tv_sec  = 1;    /* 0,5 Sekunden geht als ganzahliger Sekundenwert nur so */
			timerec.tv_usec = 500000L; /* 0,5 Sekunden in Mikrosekunden */ 
			
			uhrzeit = clock() + (clock_t) (timeout * CLK_TCK / 1000L);
			
			while (clock() < uhrzeit)
			{
				ret = win_select(NULL, readfds, writefds, NULL, &timerec);
				
				if (ret)
					break;
				for (i = 0; i < 100; i++)
					Syield();
			}
			
		}
		else
		{
			timerec.tv_sec  = (long)(timeout /1000L);    /* calc seconds from milliseconds */
			timerec.tv_usec = ((timeout%1000L)*1000L); /* calc remainder in microsecs */ 
			ret = win_select(NULL, readfds, writefds, NULL, &timerec);
		}
*/		
		if (ret > 0)
		{
			if (rfds)
			{
				memset((char*)(rfds), (char) 0, sizeof(*(rfds)));
				
				for (i = 0; i < readfds->fd_count; i++) /* !!! Achtung: Evtl. falsch; zhlt fd_count selbst mit? */
				{
					currmerk = get_entry(readfds->fd_array[i], 0);
					
					if (currmerk)
					{
						mask = 1 << currmerk->h;
						*rfds = *rfds | mask;
					}
				}
			}
			
			if (wfds)
			{
				memset((char*)(wfds), (char) 0, sizeof(*(wfds)));
				
				for (i = 0; i < writefds->fd_count; i++) /* !!! Achtung: Evtl. falsch; zhlt fd_count selbst mit? */
				{
					currmerk = get_entry(writefds->fd_array[i], 0);
					
					if (currmerk)
					{
						mask = 1 << currmerk->h;
						*wfds = *wfds | mask;
					}
				}
			}			
		}
		else
		{
			if (!ret)
			{
				add_statemsg("Timeout beim Selectaufruf!\n");
			}
			else
			{
				add_statemsg("Fehler beim Selectaufruf!\n");
				ret = -37L; /* EIHNDL - falsches Handle -- oder anderer Fehler */
			}
		}
		
		Mfree(readfds);
		Mfree(writefds);
	}
	else
	{
		add_statemsg("Fehler beim Select-Aufruf: Zuwenig Speicher!\n");
		ret = -39L; /* ENSMEM - zu wenig Speicher */
	}
	
	return ret;
}

/*============================================================================*/

long cdecl myovl_free(void)
{
	/* MERKE	*currmerk; */
	
	/* ovl_free - De-initialization of ovl (freeing memory, closing files, etc...),
		returns 0 on success, < 0 if an error has occured */
	
	/* Fr alle noch existierenden currmerk^ win_sclose(sptr) aufrufen;
		danach currmerk freigeben */

	free_merklist();

	if (old_term)
		Setexc(0x102, (void_fun_pointer) old_term);	
	
	return 0L;
}

/*============================================================================*/

void * cdecl myovl_getftab(void)
{
	static INET_FTAB inet = {
		ovl_host_addr,
		ovl_connect,
		ovl_send,
		ovl_recv,
		ovl_close,
		ovl_instat,
		ovl_select,
		ovl_info
	};

	return &inet;
}					

/*============================================================================*/

OVL_METH ovl_methods = {OVL_MAGIC, OVL_REVISION, OF_SIMPLE, FTAB_NETWORK,
								myovl_init, myovl_version, myovl_getftab, myovl_free,
								&ovl_methods.magic};

/*============================================================================*/

int main(void)
{
	if (!ovl_methods.flags) { /* OF_SIMPLE */
		exit (1); /* this shouldn't happen, but not fatal */
	
	} else if (ovl_methods.flags >= OF_THREAD) {
		exit (-EINVAL); /* mode yet not supported */
	
	} else if (ovl_methods.flags == OF_CHLDPRC) {
		if (Psigsetmask (0) == -EINVFN) {
			exit (1); /* no signal handling, notify about that */
		}
		Pause(); /* else suspend */
	}
	
	/* else OF_STARTUP, simply quit */
	return 0;
}

